(*
    This file is part of ML::AbstractUI - an abstract user interface
    API with a crossplatform implementation.

    Copyright (C) 1999 ANOQ of the Sun (alias Johnny Andersen).
    
    E-mail: anoq@HardcoreProcessing.com

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    As a special exception, if you do not do anything which is not in
    the spirit of the GNU Library General Public License, you are not
    required to physically compile this software into a separate library,
    since this is generally not possible with current Stanard ML compilers.
    However if you do something which is not in the spirit of the
    GNU Library General Public License you will have to follow the
    licence perpetually - thus disallowing you to use it for any
    commercial purposes at all.

    If you are interested in a warranty or commercial support for this
    software, contact Hardcore Processing <sales@HardcoreProcessing.com>
    for more information.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*)

require "uiregister";
require "uiwinnativenode";
require "uiwindowscore";
require "uicore";
require "parameternodes";
require "winparameternodes";
require "$.winsys.__windows_gui";
require "$.gtwinnt.dgnodes";
require "$.gtwinnt.dgpcommon";
require "uiprotocols";
require "pnotifiers";
require "lfnhierarchy";
(* FIXME: For setting up double click timer *)
require "$.basis.__timer";
require "$.basis.__time";
(* FIXME: For printing debug info of wparam and lparam *)
require "$.basis.__int";
require "windowsgeneral";

functor uiFWinListBox(Reg : UI_REGISTER
                      where type t = uiRegGUIClass.t) : UI_PLUGIN =
  struct
    structure w = WindowsGui
    val name = "ListBox"

    fun construct () =
          let
            val textlines = "TextLines"
            val selectableItems = "SelectableItems"
            val doubleClick = "OnDoubleClick"
            val click = "OnClick"
            val activate = "OnActivate"
            val deactivate = "OnDeactivate"

            val (nameNode, _) = ParamName.create name

            (* FIXME: Notifications should really be done through the connections -
                      unless it is considered illegal for the user to connect
                      stuff to size, text, create, native etc. 
                      Maybe I should just add connection constraints?
                      Or notifications for stateinfo nodes? *)

            (* Add position and size parameter nodes to the graph *)
            val (protSizeX, protSizeY) =
                  ParamSize.addXY nameNode 150 150
            val (protPosX, protPosY) =
                  ParamPos.addXY nameNode 0 0
            val visibilityProt =
                  DGPBool.add nameNode "Visible" (true, nameNode)
            val textlinesProt =
                  DGPStrings.add nameNode textlines (nil, nameNode)
            val selectableItemsProt =
                  DGPItemSelections.add nameNode selectableItems nil
            val clickNProt =
                  DGPPointerButtonNotify.add nameNode click ()
            val doubleClickNProt =
                  DGPPointerButtonNotify.add nameNode doubleClick ()
            val hasFocusProt =
                  DGPBool.add nameNode "HasFocus" (false, nameNode)
            val pointerDownNProt =
                  DGPPointerButtonNotify.add
                    nameNode "OnPointerDown" ()

            fun overrideTextLines nativeHandle =
                  let
                    fun insertStringAt index value =
                          (uiWinUtil.insertString
                             nativeHandle value index;
                           PStrings.doNotifyChanged
                             textlinesProt nameNode)

                    fun addString value =
                          (uiWinUtil.addString
                             nativeHandle value;
                           PStrings.doNotifyChanged
                             textlinesProt nameNode)

                    fun deleteStringAt index =
                          (uiWinUtil.deleteString
                             nativeHandle index;
                           PStrings.doNotifyChanged
                             textlinesProt nameNode)

                    fun countStrings () =
                          uiWinUtil.count nativeHandle

                    fun clearStrings () =
                          (uiWinUtil.clear nativeHandle;
                           PStrings.doNotifyChanged
                             textlinesProt nameNode)
                  in
                    PStrings.replaceImpl textlinesProt
                       {add = addString,
                        insertAt = insertStringAt,
                        remove = fn () => (), (* Remember notify! *)
                        deleteAt = deleteStringAt,
                        writeAt = fn index => fn value => (), (* Remember notify! *)
                        count = countStrings,
                        readAt = fn index => "",
                        clear = clearStrings,
                        readList = fn () => nil}
                  end

            fun overrideSelectableItems nativeHandle =
                  let
                    fun isSelectedAt index =
                          uiWinUtil.isSelected
                            nativeHandle index

                    fun countSelected () =
                          uiWinUtil.countSelected
                            nativeHandle
                  in
                    PItemSelections.replaceImpl selectableItemsProt
                      {isSelectedAt = isSelectedAt,
                       count = countSelected,
                       selectAt = fn _ => (),
                       deselectAll = fn () => ()}
                  end

            fun nativeCreate () =
                    let
                      val nativeProt =
                              uiWindowsCore.createWindowsWidget
                                name "LISTBOX"
                                [w.WS_VSCROLL, w.LBS_HASSTRINGS,
                                 w.LBS_MULTIPLESEL, w.LBS_EXTENDEDSEL,
                                 w.LBS_NOTIFY, (* FIXME: ??? *)
                                 w.WS_CHILD, w.WS_BORDER,
                                 w.LBS_USETABSTOPS]
                                true nameNode

                            (* FIXME: No accumulated textlines added *)
                            val nativeHandle = uiWinNative.getNativeHandle nameNode

                            val topNode =
                                  LFNHierarchy.firstParentOfType
                                    nameNode "Window"
                            val nativeTop = uiWinNative.getNativeHandle topNode

                            val sglClick = (fn _ => (PPointerButtonNotify.doNotify
                                                       pointerDownNProt nameNode
                                                       {button = 1,
                                                        point = {x = 0, y = 0}};
                                                     PPointerButtonNotify.doNotify
                                                       clickNProt nameNode
                                                       {button = 1,
                                                        point = {x = 0, y = 0}}))
                            val dblClick = (fn _ => (PPointerButtonNotify.doNotify
                                                       pointerDownNProt nameNode
                                                       {button = 1,
                                                        point = {x = 0, y = 0}};
                                                     PPointerButtonNotify.doNotify
                                                       doubleClickNProt nameNode
                                                       {button = 1,
                                                        point = {x = 0, y = 0}}))
                            
                            (* FIXME: I can't get the "real" native double click to work *)
                            (* FIXME: Encapsulate double-click timing... *)
                            val dblClickDelay = 250
                            val timer = ref NONE : Timer.real_timer option ref
                            fun realClick () =
                                  case !timer of
                                    NONE => 
                                      (timer := SOME(Timer.startRealTimer ());
                                       sglClick ())
                                  | SOME(rt) =>
                                      if Time.toMilliseconds (Timer.checkRealTimer rt) >
                                           dblClickDelay then
                                        (timer := SOME(Timer.startRealTimer ());
                                         sglClick ())
                                      else
                                        (timer := NONE; dblClick ())

                            (*val _ = w.addMessageHandler
                                      (nativeTop, w.WM_LBUTTONDOWN(*w.WM_LBUTTONDBLCLK*),
                                       fn _ => (realClick ();
                                                SOME w.nullWord))*)
                            
                            (*val _ = w.addMessageHandler
                                      (nativeHandle, w.WM_LBUTTONDOWN(*w.WM_LBUTTONDBLCLK*),
                                       fn _ => (realClick ();
                                                SOME w.nullWord))*)

                            (* FIXME: Encapsulate debugprinting of lparam and wparam *)
                            fun printWordParam param paramName =
                                  let
                                    val lol = w.loword param
                                    val hil = w.hiword param
                                    val _ = print ("High word of " ^ paramName ^
                                                   ": " ^ (Int.toString hil))
                                    val _ = print ("Low word of " ^ paramName ^
                                                   ": " ^ (Int.toString lol))
                                  in () end

                            fun printParams (w.WPARAM(wparam), w.LPARAM(lparam)) =
                                  (printWordParam wparam "wparam";
                                   printWordParam lparam "lparam")

                            fun printCommParams (_, value) =
                                  (print (Int.toString value))

                            val nativeID = uiWinNative.getNativeID nameNode

                            val _ = w.addCommandHandler
                                       (nativeTop, nativeID,
                                        fn (_, cc) =>
                                          (case cc of
                                               (* Officially: -2 = LBN_ERRSPACE *)
                                             1 => sglClick () (* Officially: LBN_SELCHANGE
                                                    (and it triggers a Click event in Delphi...) *)
                                           | 2 => dblClick () (* Officially: LBN_DBLCLK *)
                                               (* Officially: 3 = LBN_SELCANCEL *)
                                                  (* FIXME: Also override hasFocus (get/set) *)
                                           | 4 => (* Officially: LBN_SETFOCUS *)
                                                  PBool.write hasFocusProt true
                                           | 5 => (* Officially: LBN_KILLFOCUS *)
                                                  PBool.write hasFocusProt false
                                           | unknownCC => print
                                               ("Notice: Something interesting has been " ^
                                                "discovered: An unknown CommandConstant code " ^
                                                "in a ListBox CommandHandler! " ^
                                                "The code is: " ^ (Int.toString unknownCC))))

                            val _ = WindowsGeneral.overrideSizePos
                                      nativeHandle
                                      {offsetX = 0, offsetY = 0,
                                       sizeExtraX = 0, sizeExtraY = 0}
                                      {protPosX = protPosX, protPosY = protPosY,
                                       protSizeX = protSizeX, protSizeY = protSizeY}

                            val _ = overrideTextLines nativeHandle
                            val _ = overrideSelectableItems nativeHandle
                            val _ = WinParamVisibility.override visibilityProt nativeHandle

                            (*val _ = w.addMessageHandler
                                      (nativeHandle, w.LBN_DBLCLK,
                                       fn _ => (dblClick ();
                                                SOME w.nullWord))*)
                  in
                    ()
                  end

            val _ = ParamCreate.setup nameNode nativeCreate (fn _ => ())
          in
            nameNode
          end

    val _ = Reg.register name construct
  end
