(*
    This file is part of ML::AbstractUI - an abstract user interface
    API with a crossplatform implementation.

    Copyright (C) 1999 ANOQ of the Sun (alias Johnny Andersen).
    
    E-mail: anoq@HardcoreProcessing.com

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    As a special exception, if you do not do anything which is not in
    the spirit of the GNU Library General Public License, you are not
    required to physically compile this software into a separate library,
    since this is generally not possible with current Stanard ML compilers.
    However if you do something which is not in the spirit of the
    GNU Library General Public License you will have to follow the
    licence perpetually - thus disallowing you to use it for any
    commercial purposes at all.

    If you are interested in a warranty or commercial support for this
    software, contact Hardcore Processing <sales@HardcoreProcessing.com>
    for more information.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*)

require "uiregister";
require "uiwinnativenode";
require "uiwindowscore";
require "uicore";
require "parameternodes";
require "winparameternodes";
require "$.winsys.__windows_gui";
require "$.gtwinnt.dgnodes";
require "$.gtwinnt.dgpcommon";
require "pnotifiers";
require "lfnhierarchy";
require "$.basis.__int";
require "$.commonfunctions.util";
require "windowsgeneral";

functor uiFWinToggleButton(Reg : UI_REGISTER
                           where type t = uiRegGUIClass.t) : UI_PLUGIN =
  struct
    structure w = WindowsGui

    fun constructNamed name =
          let
            val text = "Text"

            val (nameNode, _) = ParamName.create name

            (* FIXME: Notifications should really be done through the connections -
                      unless it is considered illegal for the user to connect
                      stuff to size, text, create, native etc. 
                      Maybe I should just add connection constraints?
                      Or notifications for stateinfo nodes? *)

            (* Add position and size parameter nodes to the graph *)
            val (protSizeX, protSizeY) =
                  ParamSize.addXY nameNode 100 18
            val (protPosX, protPosY) =
                  ParamPos.addXY nameNode 0 0
            val visibilityProt =
                  DGPBool.add nameNode "Visible" (true, nameNode)
            val textProt =
                  DGPString.add nameNode text ("", nameNode)
            val toggleValueProt =
                  DGPBool.add nameNode "ToggleValue" (false, nameNode)
            val hasFocusProt =
                  DGPBool.add nameNode "HasFocus" (false, nameNode)
            val pointerDownNProt =
                  DGPPointerButtonNotify.add
                    nameNode "OnPointerDown" ()

            fun nativeCreate () =
                    let
                      val nativeProt =
                              uiWindowsCore.createWindowsWidget
                                name "BUTTON" [w.WS_CHILD, w.BS_CHECKBOX]
                                true nameNode

                      (* FIXME: Use nativeProt? Or maybe nativeProt is just not needed :) *)
                      val nativeHandle = uiWinNative.getNativeHandle nameNode

                      val _ = WindowsGeneral.overrideSizePos
                                nativeHandle
                                {offsetX = 0, offsetY = 0,
                                 sizeExtraX = 0, sizeExtraY = 0}
                                {protPosX = protPosX, protPosY = protPosY,
                                 protSizeX = protSizeX, protSizeY = protSizeY}

                      val _ = WindowsGeneral.overrideText nativeHandle textProt
                      val _ = WinParamVisibility.override visibilityProt nativeHandle


                      (* We must always tell windows that the toggle value should change... *)
                      fun handleToggleValueChanged _ value =
                            case value of
                              true => uiWinUtil.setInteger nativeHandle w.BM_SETCHECK 1
                            | false => uiWinUtil.setInteger nativeHandle w.BM_SETCHECK 0
                                  (* Officially: 2 = grayed checkbox *)

                      val _ = PBool.addChangedNotifier
                                toggleValueProt
                                handleToggleValueChanged

                      (* Initialize with current value *)
                      val _ = PBool.doNotifyChanged toggleValueProt
                                nameNode (PBool.read toggleValueProt)

                       val parentNode =  LFNHierarchy.parentNode nameNode
                       val topNode = LFNHierarchy.firstParentOfType
                                       nameNode "Window"
                       
                       val nativeTop = uiWinNative.getNativeHandle topNode
                       val nativeID = uiWinNative.getNativeID nameNode

                       val _ = w.addCommandHandler
                                 (nativeTop, nativeID,
                                    fn (_, cc) =>
                                      (case cc of
                                         0 =>
                                           (PBool.write toggleValueProt
                                              (not (PBool.read toggleValueProt));
                                            PPointerButtonNotify.doNotify
                                               pointerDownNProt nameNode
                                               {button = 1, point = {x = 0, y = 0}})
                                           (* Offcially: 0 = BN_CLICKED (and it changes the Checked state in Delphi...) *)
                                       | 1 => () (* Officially: 1 = BN_PAINT *)
                                       | 2 => () (* Officially: 2 = BN_HILITE / BN_PUSHED *)
                                       | 3 => () (* Officially: 3 = BN_UNHILITE / BN_UNPUSHED *)
                                       | 4 => () (* Officially: 4 = BN_DISABLE *)
                                       | 5 => () (* Officially: 5 = BN_DOUBLECLICKED / BN_DBLCLK *)
                                              (* FIXME: Also override hasFocus (get/set) *)
                                       | 6 => (* Officially: 6 = BN_SETFOCUS  *)
                                              PBool.write hasFocusProt true
                                       | 7 => (* Officially: 7 = BN_KILLFOCUS *)
                                              PBool.write hasFocusProt false
                                       | unknownCC => print
                                           ("Notice: Something interesting has been " ^
                                            "discovered: An unknown CommandConstant code " ^
                                            "in a CheckBox CommandHandler! " ^
                                            "The code is: " ^ (Int.toString unknownCC))))
                  in
                    ()
                  end

            val _ = ParamCreate.setup nameNode nativeCreate (fn _ => ())
          in
            nameNode
          end

    val name = "ToggleButton"
    val construct = fn () => constructNamed name

    val _ = Reg.register name construct
  end
