(*
    This file is part of ML::AbstractUI - an abstract user interface
    API with a crossplatform implementation.

    Copyright (C) 1999 ANOQ of the Sun (alias Johnny Andersen).
    
    E-mail: anoq@HardcoreProcessing.com

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    As a special exception, if you do not do anything which is not in
    the spirit of the GNU Library General Public License, you are not
    required to physically compile this software into a separate library,
    since this is generally not possible with current Stanard ML compilers.
    However if you do something which is not in the spirit of the
    GNU Library General Public License you will have to follow the
    licence perpetually - thus disallowing you to use it for any
    commercial purposes at all.

    If you are interested in a warranty or commercial support for this
    software, contact Hardcore Processing <sales@HardcoreProcessing.com>
    for more information.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*)

require "$.compmlworkswinnt.compmlworks";
require "$.gtwinnt.dgnodes";
require "$.gtwinnt.dgshorthand";
require "$.gtwinnt.dgpcommon";
require "uiwinnativenode";
require "parameternodes";
require "$.winsys.__windows_gui";
require "uicore_sig";
require "lfnhierarchy";

signature UI_WIN_UTIL =
  sig
    val setText : WindowsGui.hwnd -> string -> unit
    val getText : WindowsGui.hwnd -> string
    val setBuiltinFont : WindowsGui.hwnd -> WindowsGui.stock_object -> unit
    val addString : WindowsGui.hwnd -> string -> unit
    val insertString : WindowsGui.hwnd -> string -> int -> unit
    val addStrings : WindowsGui.hwnd -> StringVector.vector -> unit
    val deleteString : WindowsGui.hwnd -> int -> unit
    val count : WindowsGui.hwnd -> int
    val clear : WindowsGui.hwnd -> unit

    val countSelected : WindowsGui.hwnd -> int
    val isSelected : WindowsGui.hwnd -> int -> bool

    (* A more general style, which is preferred *)
    val setInteger : WindowsGui.hwnd -> WindowsGui.message -> int -> unit
  end

functor uiFWinUtil() : UI_WIN_UTIL =
  struct
    structure w = WindowsGui
  
    val sendMessageNoResult = ignore o w.sendMessage

    fun sendStrMsg widget str code wparam =
          let
            val string_word = w.makeCString (str)
            (* FIXME: Finalize... *)
            val result = (w.sendMessage (widget, code,
                                        wparam,
                                        w.LPARAM string_word))
                            handle e => (w.free string_word;raise e)
            val _ = w.free string_word
          in
            w.wordToInt result (* FIXME: Use wordToSignedInt to avoid overflow? *)
          end

    fun setText widget str =
          ignore (sendStrMsg widget str w.WM_SETTEXT (w.WPARAM (w.nullWord)))

    local
      fun setIt widget fontObj =
            sendMessageNoResult
              (widget,
               w.WM_SETFONT,
               w.WPARAM fontObj,
               w.LPARAM (w.intToWord 1))
    in
      fun setBuiltinFont widget font =
          let
            val w.OBJECT textFont = w.getStockObject font
          in
            setIt widget textFont
          end
            handle w.WindowSystemError _ => ()
    end

    fun text_size text =
      w.wordToSignedInt 
        (w.sendMessage (text,
                        w.WM_GETTEXTLENGTH,
                        w.WPARAM w.nullWord,
                        w.LPARAM w.nullWord))

    fun getText text =
      let
        val size = text_size text
        val buffer = w.malloc (size+1) (* extra for null termination *)
        val _ = w.sendMessage (text,
                               w.WM_GETTEXT,
                               w.WPARAM (w.intToWord (size+1)),
                               w.LPARAM buffer)
        val _ = w.setByte (buffer,size,0) (* null terminate *)
        val result = w.wordToString buffer
      in
        w.free buffer;
        result
      end

    fun addString widget str =
          ignore (sendStrMsg widget str w.LB_ADDSTRING (w.WPARAM (w.intToWord 0)))

    fun insertString widget str index =
          ignore (sendStrMsg widget str w.LB_INSERTSTRING (w.WPARAM (w.intToWord index)))

    fun addStrings widget strVector =
          let
            fun foldFnc (strLine, ()) =
                  insertString widget strLine 0
          in
            StringVector.foldl foldFnc () strVector
          end

    fun deleteString widget index =
          ignore (sendStrMsg widget "" w.LB_DELETESTRING (w.WPARAM (w.intToWord index)))

    fun count widget =
          sendStrMsg widget "" w.LB_GETCOUNT (w.WPARAM (w.nullWord))

    fun clear widget =
          ignore (sendStrMsg widget "" w.LB_RESETCONTENT (w.WPARAM (w.nullWord)))


    fun countSelected widget =
           sendStrMsg widget "" w.LB_GETSELCOUNT (w.WPARAM (w.nullWord))

    fun isSelected widget index =
          let
            val string_word = w.makeCString ("")
            (* FIXME: Finalize... *)
            val result = (w.sendMessage (widget, w.LB_GETSEL,
                                        w.WPARAM (w.intToWord index),
                                        w.LPARAM string_word))
                            handle e => (w.free string_word;raise e)
            val _ = w.free string_word
          in
            (w.wordToSignedInt result) <> 0 (* Boolean? *)
          end

    fun setInteger widget command value =
          ignore(w.sendMessage
                   (widget, command,
                    w.WPARAM (w.intToWord value),
                    w.LPARAM w.nullWord))
  end

structure uiWinUtil = uiFWinUtil()

signature UI_WINDOWS_CORE =
  sig
    include UI_NATIVE_CORE

    val createWindowsWidget : string -> string ->
          WindowsGui.window_style list -> bool ->
          DGNodes.t -> uiPWinNative.t
  end

functor uiFWindowsCore() :> UI_WINDOWS_CORE =
  struct
    structure w = WindowsGui

    (* Add windows native information to the graph *)
    fun createNativeInfo winHandle id graphNode =
          let
            val nativeNode = uiWinNative.create winHandle id
            val _ = DGNodes.connect graphNode nativeNode
          in
            DGNodes.protocol uiPWinNative.P.c nativeNode
          end

    fun createWindowsWidget name winClass winStyles createID graphNode =
          let
            open DGShorthand
            val (sizeX, sizeY) = ParamSize.getXY graphNode
            val (x, y) = ParamPos.getXY graphNode
            val id = if createID then w.newControlId() else w.nullWord
           
            val parentNode = LFNHierarchy.parentNode graphNode
            val parentType = DGNodes.typeName parentNode
            val winParent = uiWinNative.getNativeHandle parentNode

            (* FIXME: Window borders... put this in the Window control... *)
            val (sizeX, sizeY, x, y) =
                   if parentType = "MasterNode" then
                     (sizeX + 8, sizeY + 24, x - 4, y - 23)
                   else
                     (sizeX, sizeY, x, y)

            val widget =
              w.createWindowEx
                {class = winClass,
                 name = name,
                 x = x,
                 y = y,
                 width = sizeX,
                 height = sizeY,
                 parent = winParent,
                 menu = id,
                 styles = winStyles,
                 ex_styles = [w.WS_EX_STATICEDGE]}

            val _ = uiWinUtil.setBuiltinFont widget w.ANSI_VAR_FONT

            (* FIXME: Reuse visibility code here... *)
            val _ = if PBool.read (DGPBool.p (graphNode +: "Visible")) then
                         (w.showWindow (widget,w.SW_SHOWNORMAL);
                          w.updateWindow widget)
                    else ()
          in
            (* Add windows native information to the graph and return the native protocol *)
            createNativeInfo widget id graphNode
          end

    fun createMasterNode () =
          let
            val (graphNode, _) = ParamName.create "MasterNode"

            val winHandle = w.mainInit ();

            (* Add windows native information to the graph *)
            val _ = createNativeInfo winHandle w.nullWord graphNode
          in
            graphNode
          end

    fun runGUI masterNode =
          w.mainLoop ()
             handle _ => ()

    fun runApp f = f ()
  end

structure uiWindowsCore = uiFWindowsCore()
